@@ -1,6 +1,8 @@
 from pydantic import BaseModel, Field
-from typing import Optional, List
+from typing import Optional, List, Literal, Dict
 import httpx
+import asyncio
+
 from altscore.borrower_central.helpers import build_headers
 from altscore.borrower_central.model.identities import IdentitySync, IdentityAsync
 from altscore.borrower_central.model.documents import DocumentSync, DocumentAsync
@@ -11,13 +13,14 @@ from altscore.borrower_central.model.authorizations import AuthorizationSync, Au
 from altscore.borrower_central.model.relationships import RelationshipSync, RelationshipAsync
 
 from altscore.borrower_central.model.stages import StageSync, StageAsync
+from altscore.borrower_central.model.steps import StepSync, StepAsync
 from altscore.borrower_central.model.risk_ratings import RiskRatingSync, RiskRatingAsync
 from altscore.borrower_central.model.policy_alerts import AlertSync, AlertAsync
 
-from altscore.common.http_errors import raise_for_status_improved, retry_on_401
+from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
 from altscore.borrower_central.model.store_packages import PackageSync, PackageAsync
 from altscore.borrower_central.model.executions import ExecutionSync, ExecutionAsync
-from altscore.borrower_central.utils import clean_dict
+from altscore.borrower_central.utils import clean_dict, convert_to_dash_case
 
 
 class BorrowerAPIDTO(BaseModel):
@@ -25,6 +28,7 @@ class BorrowerAPIDTO(BaseModel):
     persona: str = Field(alias="persona")
     avatar_url: Optional[str] = Field(alias="avatarUrl")
     label: Optional[str] = Field(alias="label")
+    flag: Optional[str] = Field(alias="flag", default=None)
     tags: List[str] = Field(alias="tags", default=[])
     cms_client_ids: Optional[List[str]] = Field(alias="cmsClientIds", default=[])
     created_at: str = Field(alias="createdAt")
@@ -36,9 +40,74 @@ class BorrowerAPIDTO(BaseModel):
         allow_population_by_alias = True
 
 
+class SimplifiedIdentity(BaseModel):
+    id: str = Field(alias="id")
+    key: str = Field(alias="key")
+    label: str = Field(alias="label")
+    value: str = Field(alias="value")
+    priority: Optional[int] = Field(alias="priority")
+
+    class Config:
+        populate_by_name = True
+        allow_population_by_field_name = True
+        allow_population_by_alias = True
+
+
+class SimplifiedField(BaseModel):
+    id: str = Field(alias="id")
+    key: str = Field(alias="key")
+    label: str = Field(alias="label")
+    value: str = Field(alias="value")
+
+    class Config:
+        populate_by_name = True
+        allow_population_by_field_name = True
+        allow_population_by_alias = True
+
+
+class SimplifiedPointOfContact(BaseModel):
+    id: str = Field(alias="id")
+    signatures: Optional[List[str]] = Field(alias="signatures")
+    priority: int = Field(alias="priority"),
+    value: str = Field(alias="value")
+
+    class Config:
+        populate_by_name = True
+        allow_population_by_field_name = True
+        allow_population_by_alias = True
+
+
+class CurrentStep(BaseModel):
+    order: int = Field(alias="order")
+    key: str = Field(alias="key")
+    label: str = Field(alias="label")
+
+
+class BorrowerSummaryAPIDTO(BaseModel):
+    id: str = Field(alias="id")
+    persona: str = Field(alias="persona")
+    label: Optional[str] = Field(alias="label")
+    flag: Optional[str] = Field(alias="flag")
+    identities: Optional[List[SimplifiedIdentity]] = Field(alias="identities", default=[])
+    fields: Optional[List[SimplifiedField]] = Field(alias="fields", default=[])
+    points_of_contact: Optional[List[SimplifiedPointOfContact]] = Field(alias="pointsOfContact", default=[])
+    tags: List[str] = Field(alias="tags", default=[])
+    stage: Optional[str] = Field(alias="stage", default=None)
+    current_step: Optional[CurrentStep] = Field(alias="currentStep", default=None)
+    risk_rating: Optional[str] = Field(alias="riskRating", default=None)
+    created_at: str = Field(alias="createdAt")
+    updated_at: Optional[str] = Field(alias="updatedAt")
+
+    class Config:
+        populate_by_name = True
+        allow_population_by_field_name = True
+        allow_population_by_alias = True
+
+
 class CreateBorrowerDTO(BaseModel):
     persona: str = Field(alias="persona")
     label: Optional[str] = Field(alias="label")
+    flag: Optional[str] = Field(alias="flag", default=None)
     tags: List[str] = Field(alias="tags", default=[])
 
     class Config:
@@ -49,6 +118,7 @@ class CreateBorrowerDTO(BaseModel):
 
 class UpdateBorrowerDTO(BaseModel):
     label: Optional[str] = Field(alias="label", default=None)
+    flag: Optional[str] = Field(alias="flag", default=None)
     tags: List[str] = Field(alias="tags", default=[])
 
     class Config:
@@ -225,7 +295,7 @@ class BorrowersAsyncModule:
     def build_headers(self):
         return build_headers(self)
 
-    @retry_on_401
+    @retry_on_401_async
     async def create(self, new_entity_data: dict):
         async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
             response = await client.post(
@@ -237,7 +307,7 @@ class BorrowersAsyncModule:
             raise_for_status_improved(response)
             return response.json()["id"]
 
-    @retry_on_401
+    @retry_on_401_async
     async def patch(self, resource_id: str, patch_data: dict):
         async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
             response = await client.patch(
@@ -249,7 +319,7 @@ class BorrowersAsyncModule:
             raise_for_status_improved(response)
             return await self.retrieve(response.json()["id"])
 
-    @retry_on_401
+    @retry_on_401_async
     async def delete(self, resource_id: str):
         async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
             response = await client.delete(
@@ -260,7 +330,7 @@ class BorrowersAsyncModule:
             raise_for_status_improved(response)
             return None
 
-    @retry_on_401
+    @retry_on_401_async
     async def retrieve(self, resource_id: str):
         async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
             response = await client.get(
@@ -278,7 +348,7 @@ class BorrowersAsyncModule:
                 )
             return None
 
-    @retry_on_401
+    @retry_on_401_async
     async def find_one_by_identity(self, identity_key: str, identity_value: str):
         """
         Exact match by identity
@@ -304,6 +374,49 @@ class BorrowersAsyncModule:
                         return await self.retrieve(identities_found_data[0]["borrowerId"])
             return None
 
+    @retry_on_401_async
+    async def query(self, **kwargs):
+        query_params = {}
+        for k, v in kwargs.items():
+            if v is not None:
+                query_params[convert_to_dash_case(k)] = v
+
+        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
+            response = await client.get(
+                f"/v1/borrowers",
+                headers=self.build_headers(),
+                params=query_params,
+                timeout=30
+            )
+            raise_for_status_improved(response)
+            return [BorrowerAsync(
+                base_url=self.altscore_client._borrower_central_base_url,
+                header_builder=self.build_headers,
+                renew_token=self.renew_token,
+                data=BorrowerAPIDTO.parse_obj(e)
+            ) for e in response.json()]
+
+    @retry_on_401_async
+    async def query_summary(self, by: Optional[Literal["self", "identity"]] = None, search: Optional[str] = None,
+                            **kwargs):
+        query_params = {
+            "by": by,
+            "search": search
+        }
+        for k, v in kwargs.items():
+            if v is not None:
+                query_params[convert_to_dash_case(k)] = v
+        query_params = clean_dict(query_params)
+        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
+            response = await client.get(
+                f"/v1/borrowers-summary",
+                headers=self.build_headers(),
+                params=query_params,
+                timeout=30
+            )
+            raise_for_status_improved(response)
+            return [BorrowerSummaryAPIDTO.parse_obj(e) for e in response.json()]
+
 
 class BorrowersSyncModule:
 
@@ -395,6 +508,49 @@ class BorrowersSyncModule:
                         return self.retrieve(identities_found_data[0]["borrowerId"])
             return None
 
+    @retry_on_401
+    def query(self, **kwargs):
+        query_params = {}
+        for k, v in kwargs.items():
+            if v is not None:
+                query_params[convert_to_dash_case(k)] = v
+
+        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
+            response = client.get(
+                f"/v1/borrowers",
+                headers=self.build_headers(),
+                params=query_params,
+                timeout=30
+            )
+            raise_for_status_improved(response)
+            return [BorrowerSync(
+                base_url=self.altscore_client._borrower_central_base_url,
+                header_builder=self.build_headers,
+                renew_token=self.renew_token,
+                data=BorrowerAPIDTO.parse_obj(e)
+            ) for e in response.json()]
+
+    @retry_on_401
+    def query_summary(self, by: Optional[Literal["self", "identity"]] = None, search: Optional[str] = None,
+                      **kwargs):
+        query_params = {
+            "by": by,
+            "search": search
+        }
+        for k, v in kwargs.items():
+            if v is not None:
+                query_params[convert_to_dash_case(k)] = v
+        query_params = clean_dict(query_params)
+        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
+            response = client.get(
+                f"/v1/borrowers-summary",
+                headers=self.build_headers(),
+                params=query_params,
+                timeout=30
+            )
+            raise_for_status_improved(response)
+            return [BorrowerSummaryAPIDTO.parse_obj(e) for e in response.json()]
+
 
 class BorrowerAsync(BorrowerBase):
     data: BorrowerAPIDTO
@@ -405,7 +561,20 @@ class BorrowerAsync(BorrowerBase):
         self.renew_token = renew_token
         self.data = data
 
-    @retry_on_401
+    @retry_on_401_async
+    async def set_label(self, label: str):
+        async with httpx.AsyncClient(base_url=self.base_url) as client:
+            response = await client.patch(
+                f"{self.base_url}/v1/borrowers/{self.data.id}",
+                headers=self._header_builder(),
+                json={
+                    "label": label,
+                }
+            )
+            raise_for_status_improved(response)
+            return None
+
+    @retry_on_401_async
     async def get_stage(self) -> StageAsync:
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             response = await client.get(
@@ -420,7 +589,34 @@ class BorrowerAsync(BorrowerBase):
                 data=response.json()
             )
 
-    @retry_on_401
+    @retry_on_401_async
+    async def get_current_step(self) -> StepAsync:
+        async with httpx.AsyncClient(base_url=self.base_url) as client:
+            response = await client.get(
+                f"{self.base_url}/v1/borrowers/{self.data.id}/steps/current",
+                headers=self._header_builder()
+            )
+            raise_for_status_improved(response)
+            return StepAsync(
+                base_url=self.base_url,
+                header_builder=self._header_builder,
+                renew_token=self.renew_token,
+                data=response.json()
+            )
+
+    @retry_on_401_async
+    async def set_current_step(self, key: str):
+        async with httpx.AsyncClient(base_url=self.base_url) as client:
+            response = await client.put(
+                f"{self.base_url}/v1/borrowers/{self.data.id}/steps/current",
+                json={
+                    "key": key
+                },
+                headers=self._header_builder()
+            )
+            raise_for_status_improved(response)
+
+    @retry_on_401_async
     async def set_stage(self, stage: str, reference_id: Optional[str] = None):
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             response = await client.put(
@@ -434,7 +630,7 @@ class BorrowerAsync(BorrowerBase):
             raise_for_status_improved(response)
             return None
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_risk_rating(self) -> RiskRatingAsync:
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             response = await client.get(
@@ -449,7 +645,7 @@ class BorrowerAsync(BorrowerBase):
                 data=response.json()
             )
 
-    @retry_on_401
+    @retry_on_401_async
     async def set_risk_rating(self, risk_rating: str, reference_id: Optional[str] = None):
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             response = await client.put(
@@ -463,7 +659,7 @@ class BorrowerAsync(BorrowerBase):
             raise_for_status_improved(response)
             return None
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_documents(self, **kwargs) -> List[DocumentAsync]:
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             url, query = self._documents(self.data.id, **kwargs)
@@ -480,7 +676,7 @@ class BorrowerAsync(BorrowerBase):
                 data=document_data
             ) for document_data in data]
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_identity_by_key(self, key: str) -> Optional[IdentityAsync]:
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             url, query = self._identities(self.data.id, key=key)
@@ -499,7 +695,7 @@ class BorrowerAsync(BorrowerBase):
                 data=data[0]
             )
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_borrower_field_by_key(self, key: str) -> Optional[BorrowerFieldAsync]:
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             url, query = self._borrower_fields(self.data.id, key=key)
@@ -518,7 +714,7 @@ class BorrowerAsync(BorrowerBase):
                 data=data[0]
             )
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_identities(self, **kwargs) -> List[IdentityAsync]:
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             url, query = self._identities(self.data.id, **kwargs)
@@ -535,7 +731,7 @@ class BorrowerAsync(BorrowerBase):
                 data=identity_data
             ) for identity_data in data]
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_addresses(self, **kwargs) -> List[AddressAsync]:
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             url, query = self._addresses(self.data.id, **kwargs)
@@ -552,7 +748,7 @@ class BorrowerAsync(BorrowerBase):
                 data=address_data
             ) for address_data in data]
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_points_of_contact(self, **kwargs) -> List[PointOfContactAsync]:
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             url, query = self._points_of_contact(self.data.id, **kwargs)
@@ -569,7 +765,7 @@ class BorrowerAsync(BorrowerBase):
                 data=point_of_contact_data
             ) for point_of_contact_data in data]
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_borrower_fields(self, **kwargs) -> List[BorrowerFieldAsync]:
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             url, query = self._borrower_fields(self.data.id, **kwargs)
@@ -586,7 +782,7 @@ class BorrowerAsync(BorrowerBase):
                 data=borrower_field_data
             ) for borrower_field_data in data]
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_authorizations(self, **kwargs) -> List[AuthorizationAsync]:
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             url, query = self._authorizations(self.data.id, **kwargs)
@@ -603,7 +799,7 @@ class BorrowerAsync(BorrowerBase):
                 data=authorization_data
             ) for authorization_data in data]
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_relationships(self, **kwargs) -> List[RelationshipAsync]:
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             url, query = self._relationships(self.data.id, **kwargs)
@@ -620,7 +816,7 @@ class BorrowerAsync(BorrowerBase):
                 data=relationship_data
             ) for relationship_data in data]
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_executions(self, **kwargs) -> List[ExecutionAsync]:
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             url, query = self._executions(self.data.id, **kwargs)
@@ -637,7 +833,7 @@ class BorrowerAsync(BorrowerBase):
                 data=execution_data
             ) for execution_data in data]
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_packages(self, **kwargs) -> List[PackageAsync]:
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             url, query = self._packages(self.data.id, **kwargs)
@@ -654,7 +850,7 @@ class BorrowerAsync(BorrowerBase):
                 data=package_data
             ) for package_data in data]
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_alerts(self, **kwargs) -> List[AlertAsync]:
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             url, query = self._alerts(self.data.id, **kwargs)
@@ -672,7 +868,7 @@ class BorrowerAsync(BorrowerBase):
             ) for alert_data in data]
             return alerts
 
-    @retry_on_401
+    @retry_on_401_async
     async def associate_cms_client_id(self, cms_client_id: str):
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             response = await client.post(
@@ -682,7 +878,7 @@ class BorrowerAsync(BorrowerBase):
             raise_for_status_improved(response)
             return None
 
-    @retry_on_401
+    @retry_on_401_async
     async def put_cms_client_ids(self, cms_client_ids: List[str]):
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             response = await client.put(
@@ -695,7 +891,7 @@ class BorrowerAsync(BorrowerBase):
             raise_for_status_improved(response)
             return None
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_main_address(self) -> Optional[AddressAsync]:
         addresses = await self.get_addresses(sort_by="priority", per_page=1)
         if len(addresses) == 0:
@@ -703,7 +899,7 @@ class BorrowerAsync(BorrowerBase):
         else:
             return addresses[0]
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_main_point_of_contact(self, contact_method: str) -> Optional[PointOfContactAsync]:
         points_of_contact = await self.get_points_of_contact(
             contact_method=contact_method, sort_by="priority", per_page=1
@@ -713,6 +909,35 @@ class BorrowerAsync(BorrowerBase):
         else:
             return points_of_contact[0]
 
+    @retry_on_401
+    async def map_identities_and_fields_onto_dict(self, mapping_dict: dict):
+        identities_to_query = {k: 1 for k in mapping_dict.keys() if k.startswith("identity.")}
+        malformed_identities = [k for k in identities_to_query if len(k.split(".")[-1]) == 0]
+        borrower_fields_to_query = {k: 1 for k in mapping_dict.keys() if k.startswith("borrower_field.")}
+        malformed_fields = [k for k in borrower_fields_to_query if len(k.split(".")[-1]) == 0]
+        if len(malformed_fields + malformed_identities) > 0:
+            raise ValueError(f"Found malformed keys: {malformed_fields + malformed_identities}")
+        calls = []
+        for identity_key in identities_to_query:
+            calls.append(
+                self.get_identity_by_key(identity_key)
+            )
+        for field_key in borrower_fields_to_query:
+            calls.append(
+                self.get_borrower_field_by_key(field_key)
+            )
+        value_maps = {}
+        calls = await asyncio.gather(*calls)
+        for element in calls:
+            if identities_to_query.get(element.data.key) is not None:
+                value_maps[f"identity.{element.data.key}"] = element.data.value
+            elif borrower_fields_to_query.get(element.data.key) is not None:
+                value_maps[f"borrower_field.{element.data.key}"] = element.data.value
+        mapped_dict = {}
+        for k, v_map in mapping_dict:
+            mapped_dict[k] = value_maps.get(v_map)
+        return mapped_dict
+
     def __str__(self):
         return str(self.data)
 
@@ -729,6 +954,19 @@ class BorrowerSync(BorrowerBase):
         self.renew_token = renew_token
         self.data = data
 
+    @retry_on_401
+    def set_label(self, label: str):
+        with httpx.Client(base_url=self.base_url) as client:
+            response = client.patch(
+                f"{self.base_url}/v1/borrowers/{self.data.id}",
+                headers=self._header_builder(),
+                json={
+                    "label": label,
+                }
+            )
+            raise_for_status_improved(response)
+            return None
+
     @retry_on_401
     def get_stage(self):
         with httpx.Client(base_url=self.base_url) as client:
@@ -744,6 +982,33 @@ class BorrowerSync(BorrowerBase):
                 data=response.json()
             )
 
+    @retry_on_401
+    def get_current_step(self) -> StepSync:
+        with httpx.Client(base_url=self.base_url) as client:
+            response = client.get(
+                f"{self.base_url}/v1/borrowers/{self.data.id}/steps/current",
+                headers=self._header_builder()
+            )
+            raise_for_status_improved(response)
+            return StepSync(
+                base_url=self.base_url,
+                header_builder=self._header_builder,
+                renew_token=self.renew_token,
+                data=response.json()
+            )
+
+    @retry_on_401
+    def set_current_step(self, key: str):
+        with httpx.Client(base_url=self.base_url) as client:
+            response = client.put(
+                f"{self.base_url}/v1/borrowers/{self.data.id}/steps/current",
+                json={
+                    "key": key
+                },
+                headers=self._header_builder()
+            )
+            raise_for_status_improved(response)
+
     @retry_on_401
     def set_stage(self, stage: str, reference_id: Optional[str] = None):
         with httpx.Client(base_url=self.base_url) as client:
@@ -1037,6 +1302,34 @@ class BorrowerSync(BorrowerBase):
         else:
             return points_of_contact[0]
 
+    @retry_on_401
+    def map_identities_and_fields_onto_dict(self, mapping_dict: dict):
+        identities_to_query = {k: 1 for k in mapping_dict.keys() if k.startswith("identity.")}
+        malformed_identities = [k for k in identities_to_query if len(k.split(".")[-1]) == 0]
+        borrower_fields_to_query = {k: 1 for k in mapping_dict.keys() if k.startswith("borrower_field.")}
+        malformed_fields = [k for k in borrower_fields_to_query if len(k.split(".")[-1]) == 0]
+        if len(malformed_fields + malformed_identities) > 0:
+            raise ValueError(f"Found malformed keys: {malformed_fields + malformed_identities}")
+        calls = []
+        for identity_key in identities_to_query:
+            calls.append(
+                self.get_identity_by_key(identity_key)
+            )
+        for field_key in borrower_fields_to_query:
+            calls.append(
+                self.get_borrower_field_by_key(field_key)
+            )
+        value_maps = {}
+        for element in calls:
+            if identities_to_query.get(element.data.key) is not None:
+                value_maps[f"identity.{element.data.key}"] = element.data.value
+            elif borrower_fields_to_query.get(element.data.key) is not None:
+                value_maps[f"borrower_field.{element.data.key}"] = element.data.value
+        mapped_dict = {}
+        for k, v_map in mapping_dict:
+            mapped_dict[k] = value_maps.get(v_map)
+        return mapped_dict
+
     def __str__(self):
         return str(self.data)
 
