@@ -2,9 +2,9 @@ import httpx
 import json
 from altscore.borrower_central.helpers import build_headers
 from altscore.borrower_central.model.attachments import AttachmentAPIDTO, AttachmentInput
-from altscore.common.http_errors import raise_for_status_improved, retry_on_401
+from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
 from typing import Dict
-import stringcase
+from altscore.borrower_central.utils import convert_to_dash_case
 
 
 class GenericBase:
@@ -102,7 +102,7 @@ class GenericAsyncResource(GenericBase):
     def created_at(self):
         return self.data.created_at
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_attachments(self):
         if self.data.has_attachments:
             async with httpx.AsyncClient() as client:
@@ -115,7 +115,7 @@ class GenericAsyncResource(GenericBase):
                 self.attachments = [AttachmentAPIDTO.parse_obj(e) for e in response.json()]
         return self.attachments
 
-    @retry_on_401
+    @retry_on_401_async
     async def post_attachment(self, attachment: Dict):
         async with httpx.AsyncClient() as client:
             response = await client.post(
@@ -126,7 +126,7 @@ class GenericAsyncResource(GenericBase):
             )
             raise_for_status_improved(response)
 
-    @retry_on_401
+    @retry_on_401_async
     async def get_content(self):
         if self.resource in ["stores/packages"]:
             async with httpx.AsyncClient() as client:
@@ -174,7 +174,40 @@ class GenericSyncModule:
                     renew_token=self.renew_token,
                     data=self.retrieve_data_model.parse_obj(response.json())
                 )
-            return None
+            elif response.status_code in [404]:
+                return None
+
+            raise_for_status_improved(response)
+
+    @retry_on_401
+    def retrieve_all(self, **kwargs):
+        query_params = {}
+        per_page = 10
+        for k, v in kwargs.items():
+            if v is not None:
+                query_params[convert_to_dash_case(k)] = v
+        query_params["per-page"] = per_page
+        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
+            response = client.get(
+                f"/v1/{self.resource}",
+                params=query_params,
+                headers=self.build_headers(),
+                timeout=30
+            )
+            raise_for_status_improved(response)
+            total_count = int(response.headers["x-total-count"])
+        resources = []
+        # TODO: this is not optimal, we should use asyncio.gather and a batch size
+        total_pages = (total_count // per_page) + 1
+        if total_pages > 1:
+            pages = range(1, total_pages + 1)
+        else:
+            pages = [1]
+        for page in pages:
+            r = self.query(page=page, per_page=per_page, **kwargs)
+            resources.append(r)
+        resources = [item for sublist in resources for item in sublist]
+        return resources
 
     @retry_on_401
     def create(self, new_entity_data: Dict, update_if_exists: bool = False) -> str:
@@ -270,7 +303,7 @@ class GenericAsyncModule:
     def print_retrieve_schema(self):
         print(json.dumps(self.retrieve_data_model.schema(), indent=2, ensure_ascii=False))
 
-    @retry_on_401
+    @retry_on_401_async
     async def retrieve(self, resource_id: str):
         async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
             response = await client.get(
@@ -285,11 +318,41 @@ class GenericAsyncModule:
                     renew_token=self.renew_token,
                     data=self.retrieve_data_model.parse_obj(response.json())
                 )
-            elif response.status_code in [403, 401]:
-                raise Exception("Unauthorized, check your API key")
-            return None
+            elif response.status_code in [404]:
+                return None
 
-    @retry_on_401
+            raise_for_status_improved(response)
+
+    @retry_on_401_async
+    async def retrieve_all(self, **kwargs):
+        query_params = {}
+        per_page = 10
+        for k, v in kwargs.items():
+            if v is not None:
+                query_params[convert_to_dash_case(k)] = v
+        query_params["per-page"] = per_page
+        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
+            response = await client.get(
+                f"/v1/{self.resource}",
+                params=query_params,
+                headers=self.build_headers(),
+                timeout=30
+            )
+            raise_for_status_improved(response)
+            total_count = int(response.headers["x-total-count"])
+        resources = []
+        # TODO: this is not optimal, we should use asyncio.gather and a batch size
+        total_pages = (total_count // per_page) + 1
+        if total_pages > 1:
+            pages = range(1, total_pages + 1)
+        else:
+            pages = [1]
+        for page in pages:
+            resources.append(await self.query(page=page, per_page=per_page, **kwargs))
+        resources = [item for sublist in resources for item in sublist]
+        return resources
+
+    @retry_on_401_async
     async def create(self, new_entity_data: Dict, update_if_exists: bool = False) -> str:
         async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
             response = await client.post(
@@ -308,7 +371,7 @@ class GenericAsyncModule:
             raise_for_status_improved(response)
             return response.json()["id"]
 
-    @retry_on_401
+    @retry_on_401_async
     async def patch(self, resource_id: str, patch_data: Dict) -> str:
         async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
             response = await client.patch(
@@ -320,7 +383,7 @@ class GenericAsyncModule:
             raise_for_status_improved(response)
             return resource_id
 
-    @retry_on_401
+    @retry_on_401_async
     async def delete(self, resource_id: str):
         async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
             response = await client.delete(
@@ -331,7 +394,7 @@ class GenericAsyncModule:
             raise_for_status_improved(response)
             return None
 
-    @retry_on_401
+    @retry_on_401_async
     async def query(self, **kwargs):
         query_params = {}
         for k, v in kwargs.items():
@@ -352,8 +415,3 @@ class GenericAsyncModule:
                 renew_token=self.renew_token,
                 data=self.retrieve_data_model.parse_obj(e)
             ) for e in response.json()]
-
-
-def convert_to_dash_case(s):
-    snake_case = stringcase.snakecase(s)
-    return stringcase.spinalcase(snake_case)
