@@ -1,7 +1,7 @@
 from pydantic import BaseModel, Field
-from typing import List
+from typing import List, Optional
 import httpx
-from altscore.common.http_errors import raise_for_status_improved, retry_on_401
+from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
 from altscore.cms.model.generics import GenericSyncModule, GenericAsyncModule
 from altscore.cms.model.common import Money, Schedule, Terms
 import datetime as dt
@@ -27,7 +27,7 @@ class DPAFlowAPIDTO(BaseModel):
     reference_id: str = Field(alias="referenceId")
     status: str = Field(alias="status")
     client: Client = Field(alias="client")
-    schedule: Schedule = Field(alias="schedule")
+    schedule: List[Schedule] = Field(alias="schedule")
     terms: Terms = Field(alias="terms")
     created_at: str = Field(alias="createdAt")
     updated_at: str = Field(alias="updatedAt")
@@ -41,7 +41,7 @@ class DPAFlowAPIDTO(BaseModel):
 class CreateDPAFlowDTO(BaseModel):
     amount: Money = Field(alias="amount")
     disbursement_date: str = Field(alias="disbursementDate")
-    external_id: str = Field(alias="externalId")
+    client_id: str = Field(alias="clientId", default=None)
 
     class Config:
         populate_by_name = True
@@ -75,6 +75,19 @@ class Invoice(BaseModel):
         populate_by_alias = True
 
 
+class ApproveDPAFlowDTO(BaseModel):
+    amount: Money = Field(alias="amount")
+    disbursement_date: str = Field(alias="disbursementDate")
+    client_id: Optional[str] = Field(alias="clientId", default=None)
+    external_id: Optional[str] = Field(alias="externalId", default=None)
+    reference_id: Optional[str] = Field(alias="referenceId", default=None)
+
+    class Config:
+        populate_by_name = True
+        allow_population_by_field_name = True
+        populate_by_alias = True
+
+
 class DPABase:
 
     @staticmethod
@@ -100,21 +113,22 @@ class DPAFlowAsync(DPABase):
         self.renew_token = renew_token
         self.data = data
 
-    @retry_on_401
-    async def approve(self):
+    @retry_on_401_async
+    async def approve(self, approve_data: dict):
         async with httpx.AsyncClient(base_url=self.base_url) as client:
-            response = await client.post(
+            response = await client.put(
                 self._approval(self.data.id),
                 headers=self._header_builder(),
+                json=ApproveDPAFlowDTO.parse_obj(approve_data).dict(by_alias=True),
                 timeout=30
             )
             raise_for_status_improved(response)
             self.data = DPAFlowAPIDTO.parse_obj(response.json())
 
-    @retry_on_401
+    @retry_on_401_async
     async def cancel(self):
         async with httpx.AsyncClient(base_url=self.base_url) as client:
-            response = await client.post(
+            response = await client.put(
                 self._cancellation(self.data.id),
                 headers=self._header_builder(),
                 timeout=30
@@ -122,12 +136,12 @@ class DPAFlowAsync(DPABase):
             raise_for_status_improved(response)
             self.data = DPAFlowAPIDTO.parse_obj(response.json())
 
-    @retry_on_401
-    async def submit_invoice(self, invoice: dict):
+    @retry_on_401_async
+    async def submit_invoice(self, invoice_data: dict):
         async with httpx.AsyncClient(base_url=self.base_url) as client:
             response = await client.post(
                 self._invoice(self.data.id),
-                json=Invoice.parse_obj(invoice).dict(by_alias=True),
+                json=Invoice.parse_obj(invoice_data).dict(by_alias=True),
                 headers=self._header_builder(),
                 timeout=30
             )
@@ -152,11 +166,12 @@ class DPAFlowSync(DPABase):
         self.data: DPAFlowAPIDTO = data
 
     @retry_on_401
-    def approve(self):
+    def approve(self, approve_data: dict):
         with httpx.Client(base_url=self.base_url) as client:
-            response = client.post(
+            response = client.put(
                 self._approval(self.data.id),
                 headers=self._header_builder(),
+                json=ApproveDPAFlowDTO.parse_obj(approve_data).dict(by_alias=True),
                 timeout=30
             )
             raise_for_status_improved(response)
@@ -165,7 +180,7 @@ class DPAFlowSync(DPABase):
     @retry_on_401
     def cancel(self):
         with httpx.Client(base_url=self.base_url) as client:
-            response = client.post(
+            response = client.put(
                 self._cancellation(self.data.id),
                 headers=self._header_builder(),
                 timeout=30
@@ -173,6 +188,18 @@ class DPAFlowSync(DPABase):
             raise_for_status_improved(response)
             self.data = DPAFlowAPIDTO.parse_obj(response.json())
 
+    @retry_on_401
+    def submit_invoice(self, invoice_data: dict):
+        with httpx.Client(base_url=self.base_url) as client:
+            response = client.post(
+                self._invoice(self.data.id),
+                headers=self._header_builder(),
+                json=Invoice.parse_obj(invoice_data).dict(by_alias=True),
+                timeout=300
+            )
+            raise_for_status_improved(response)
+            return response.json()
+
     def __str__(self):
         return str(self.data)
 
@@ -192,7 +219,7 @@ class DPAFlowsAsyncModule(GenericAsyncModule):
             resource="dpas"
         )
 
-    @retry_on_401
+    @retry_on_401_async
     async def create(self, new_entity_data: dict):
         disbursement_date_str = \
             new_entity_data.get("disbursementDate") or new_entity_data.get("disbursement_date")
@@ -212,6 +239,26 @@ class DPAFlowsAsyncModule(GenericAsyncModule):
             raise_for_status_improved(response)
             return self.retrieve_data_model.parse_obj(response.json()).id
 
+    @retry_on_401_async
+    async def simulate(self, new_entity_data: dict):
+        disbursement_date_str = \
+            new_entity_data.get("disbursementDate") or new_entity_data.get("disbursement_date")
+        if disbursement_date_str is None:
+            disbursement_date = dt.date.today().strftime("%Y-%m-%d")
+        else:
+            disbursement_date = date_parser(disbursement_date_str).strftime("%Y-%m-%d")
+        new_entity_data["disbursementDate"] = disbursement_date
+
+        async with httpx.AsyncClient(base_url=self.altscore_client._cms_base_url) as client:
+            response = await client.post(
+                f"/{self.resource_version}/{self.resource}/simulations",
+                headers=self.build_headers(),
+                json=self.create_data_model.parse_obj(new_entity_data).dict(by_alias=True),
+                timeout=30
+            )
+            raise_for_status_improved(response)
+            return self.retrieve_data_model.parse_obj(response.json())
+
 
 class DPAFlowsSyncModule(GenericSyncModule):
 
@@ -244,3 +291,23 @@ class DPAFlowsSyncModule(GenericSyncModule):
             )
             raise_for_status_improved(response)
             return self.retrieve_data_model.parse_obj(response.json()).id
+
+    @retry_on_401
+    def simuate(self, new_entity_data: dict):
+        disbursement_date_str = \
+            new_entity_data.get("disbursementDate") or new_entity_data.get("disbursement_date")
+        if disbursement_date_str is None:
+            disbursement_date = dt.date.today().strftime("%Y-%m-%d")
+        else:
+            disbursement_date = date_parser(disbursement_date_str).strftime("%Y-%m-%d")
+        new_entity_data["disbursementDate"] = disbursement_date
+
+        with httpx.Client(base_url=self.altscore_client._cms_base_url) as client:
+            response = client.post(
+                f"/{self.resource_version}/{self.resource}/simulations",
+                headers=self.build_headers(),
+                json=self.create_data_model.parse_obj(new_entity_data).dict(by_alias=True),
+                timeout=30
+            )
+            raise_for_status_improved(response)
+            return self.retrieve_data_model.parse_obj(response.json())
